@@ -60,10 +60,14 @@ fn prefer_outer_delimiter(language: guess_language::Language) -> bool {
         // JSON and TOML are like Lisp: the outer delimiter in an array object
         // is the most relevant.
         Json | Toml | Hcl => true,
+        // It's probably the case that outer delimiters are used more
+        // frequently than inner delimiters in SQl. (foo = 1 OR bar = 2)
+        // is more likely than foo(1).
+        Sql => true,
         // For everything else, prefer the inner delimiter. These
         // languages have syntax like `foo(bar)` or `foo[bar]` where
         // the inner delimiter is more relevant.
-        Bash | C | CPlusPlus | CSharp | Css | Dart | Elixir | Elm | Elvish | Gleam | Go
+        Bash | C | CPlusPlus | CSharp | Css | Dart | Elixir | Elm | Elvish | Gleam | Go | Hack
         | Haskell | Html | Java | JavaScript | Jsx | Julia | Kotlin | Lua | Nix | OCaml
         | OCamlInterface | Perl | Php | Python | Ruby | Rust | Scala | Swift | Tsx | TypeScript
         | Yaml | Zig => false,
