@@ -7,11 +7,11 @@ use std::{cmp::max, collections::HashSet};
 use crate::{
     constants::Side,
     display::context::all_matched_lines_filled,
-    display::hunks::{matched_lines_for_hunk, Hunk},
+    display::hunks::{matched_lines_indexes_for_hunk, Hunk},
     display::style::{
         self, apply_colors, color_positions, novel_style, split_and_apply, BackgroundColor,
     },
-    lines::{codepoint_len, format_line_num, LineNumber},
+    lines::{codepoint_len, format_line_num, split_on_newlines, LineNumber},
     options::{DisplayMode, DisplayOptions},
     parse::syntax::{zip_pad_shorter, MatchedPos},
     positions::SingleLineSpan,
@@ -19,22 +19,6 @@ use crate::{
 
 const SPACER: &str = " ";
 
-/// Split `s` on \n or \r\n. Always returns a non-empty vec.
-///
-/// This differs from `str::lines`, which considers `""` to be zero
-/// lines and `"foo\n"` to be one line.
-fn split_on_newlines(s: &str) -> Vec<&str> {
-    s.split('\n')
-        .map(|l| {
-            if let Some(l) = l.strip_suffix('\r') {
-                l
-            } else {
-                l
-            }
-        })
-        .collect()
-}
-
 fn format_line_num_padded(line_num: LineNumber, column_width: usize) -> String {
     format!(
         "{:width$} ",
@@ -81,14 +65,16 @@ fn display_single_column(
     lhs_display_path: &str,
     rhs_display_path: &str,
     lang_name: &str,
-    src: &str,
+    src_lines: &[String],
     is_lhs: bool,
     display_options: &DisplayOptions,
-) -> String {
-    let column_width = format_line_num(src.lines().count().into()).len();
+) -> Vec<String> {
+    let column_width = format_line_num((src_lines.len() as u32).into()).len();
 
-    let mut result = String::with_capacity(src.len());
-    result.push_str(&style::header(
+    let mut result = Vec::with_capacity(src_lines.len());
+
+    let mut header_line = String::new();
+    header_line.push_str(&style::header(
         lhs_display_path,
         rhs_display_path,
         1,
@@ -96,21 +82,23 @@ fn display_single_column(
         lang_name,
         display_options,
     ));
-    result.push('\n');
+    header_line.push('\n');
+    result.push(header_line);
 
     let mut style = Style::new();
     if display_options.use_color {
         style = novel_style(Style::new(), is_lhs, display_options.background_color);
     }
 
-    for (i, line) in src.lines().enumerate() {
-        result.push_str(
-            &format_line_num_padded(i.into(), column_width)
+    for (i, line) in src_lines.iter().enumerate() {
+        let mut formatted_line = String::with_capacity(line.len());
+        formatted_line.push_str(
+            &format_line_num_padded((i as u32).into(), column_width)
                 .style(style)
                 .to_string(),
         );
-        result.push_str(line);
-        result.push('\n');
+        formatted_line.push_str(line);
+        result.push(formatted_line);
     }
 
     result
@@ -197,11 +185,17 @@ impl SourceDimensions {
         for (lhs_line_num, rhs_line_num) in line_nums {
             if let Some(lhs_line_num) = lhs_line_num {
                 lhs_max_line = max(lhs_max_line, *lhs_line_num);
-                lhs_max_content = max(lhs_max_content, codepoint_len(lhs_lines[lhs_line_num.0]));
+                lhs_max_content = max(
+                    lhs_max_content,
+                    codepoint_len(lhs_lines[lhs_line_num.as_usize()]),
+                );
             }
             if let Some(rhs_line_num) = rhs_line_num {
                 rhs_max_line = max(rhs_max_line, *rhs_line_num);
-                rhs_max_content = max(rhs_max_content, codepoint_len(rhs_lines[rhs_line_num.0]));
+                rhs_max_content = max(
+                    rhs_max_content,
+                    codepoint_len(rhs_lines[rhs_line_num.as_usize()]),
+                );
             }
         }
 
@@ -283,7 +277,7 @@ fn highlight_as_novel(
             return true;
         }
 
-        let line_content = lines.get(line_num.0).map(|s| str::trim(s));
+        let line_content = lines.get(line_num.as_usize()).map(|s| str::trim(s));
         // If this is a blank line without a corresponding line on the
         // other side, highlight it too. This helps highlight novel
         // blank lines.
@@ -306,7 +300,7 @@ pub fn print(
     lhs_mps: &[MatchedPos],
     rhs_mps: &[MatchedPos],
 ) {
-    let (lhs_colored_src, rhs_colored_src) = if display_options.use_color {
+    let (lhs_colored_lines, rhs_colored_lines) = if display_options.use_color {
         (
             apply_colors(
                 lhs_src,
@@ -324,35 +318,44 @@ pub fn print(
             ),
         )
     } else {
-        (lhs_src.to_string(), rhs_src.to_string())
+        (
+            split_on_newlines(lhs_src)
+                .iter()
+                .map(|s| (*s).to_owned())
+                .collect(),
+            split_on_newlines(rhs_src)
+                .iter()
+                .map(|s| (*s).to_owned())
+                .collect(),
+        )
     };
 
     if lhs_src.is_empty() {
-        println!(
-            "{}",
-            display_single_column(
-                lhs_display_path,
-                rhs_display_path,
-                lang_name,
-                &rhs_colored_src,
-                false,
-                display_options
-            )
-        );
+        for line in display_single_column(
+            lhs_display_path,
+            rhs_display_path,
+            lang_name,
+            &rhs_colored_lines,
+            false,
+            display_options,
+        ) {
+            print!("{}", line);
+        }
+        println!();
         return;
     }
     if rhs_src.is_empty() {
-        println!(
-            "{}",
-            display_single_column(
-                lhs_display_path,
-                rhs_display_path,
-                lang_name,
-                &lhs_colored_src,
-                true,
-                display_options
-            )
-        );
+        for line in display_single_column(
+            lhs_display_path,
+            rhs_display_path,
+            lang_name,
+            &lhs_colored_lines,
+            true,
+            display_options,
+        ) {
+            print!("{}", line);
+        }
+        println!();
         return;
     }
 
@@ -368,17 +371,15 @@ pub fn print(
         (FxHashMap::default(), FxHashMap::default())
     };
 
-    let lhs_lines = split_on_newlines(lhs_src);
-    let rhs_lines = split_on_newlines(rhs_src);
-    let lhs_colored_lines = split_on_newlines(&lhs_colored_src);
-    let rhs_colored_lines = split_on_newlines(&rhs_colored_src);
-
     let (lhs_lines_with_novel, rhs_lines_with_novel) = lines_with_novel(lhs_mps, rhs_mps);
 
     let mut prev_lhs_line_num = None;
     let mut prev_rhs_line_num = None;
 
+    let lhs_lines = split_on_newlines(lhs_src);
+    let rhs_lines = split_on_newlines(rhs_src);
     let matched_lines = all_matched_lines_filled(lhs_mps, rhs_mps, &lhs_lines, &rhs_lines);
+    let mut matched_lines_to_print = &matched_lines[..];
 
     for (i, hunk) in hunks.iter().enumerate() {
         println!(
@@ -393,34 +394,42 @@ pub fn print(
             )
         );
 
-        let aligned_lines = matched_lines_for_hunk(&matched_lines, hunk);
+        let (start_i, end_i) = matched_lines_indexes_for_hunk(matched_lines_to_print, hunk);
+        let aligned_lines = &matched_lines_to_print[start_i..end_i];
+        // We iterate through hunks in order, so we know the next hunk
+        // must appear after start_i. This makes
+        // `matched_lines_indexes_for_hunk` faster on later
+        // iterations, and this function is hot on large textual
+        // diffs.
+        matched_lines_to_print = &matched_lines_to_print[start_i..];
+
         let no_lhs_changes = hunk.novel_lhs.is_empty();
         let no_rhs_changes = hunk.novel_rhs.is_empty();
         let same_lines = aligned_lines.iter().all(|(l, r)| l == r);
 
         let source_dims = SourceDimensions::new(
             display_options.display_width,
-            &aligned_lines,
+            aligned_lines,
             &lhs_lines,
             &rhs_lines,
         );
         for (lhs_line_num, rhs_line_num) in aligned_lines {
             let lhs_line_novel = highlight_as_novel(
-                lhs_line_num,
+                *lhs_line_num,
                 &lhs_lines,
-                rhs_line_num,
+                *rhs_line_num,
                 &lhs_lines_with_novel,
             );
             let rhs_line_novel = highlight_as_novel(
-                rhs_line_num,
+                *rhs_line_num,
                 &rhs_lines,
-                lhs_line_num,
+                *lhs_line_num,
                 &rhs_lines_with_novel,
             );
 
             let (display_lhs_line_num, display_rhs_line_num) = display_line_nums(
-                lhs_line_num,
-                rhs_line_num,
+                *lhs_line_num,
+                *rhs_line_num,
                 &source_dims,
                 display_options.use_color,
                 display_options.background_color,
@@ -437,11 +446,11 @@ pub fn print(
             if no_lhs_changes && !show_both {
                 match rhs_line_num {
                     Some(rhs_line_num) => {
-                        let rhs_line = &rhs_colored_lines[rhs_line_num.0];
+                        let rhs_line = &rhs_colored_lines[rhs_line_num.as_usize()];
                         if same_lines {
-                            println!("{}{}", display_rhs_line_num, rhs_line);
+                            print!("{}{}", display_rhs_line_num, rhs_line);
                         } else {
-                            println!(
+                            print!(
                                 "{}{}{}",
                                 display_lhs_line_num, display_rhs_line_num, rhs_line
                             );
@@ -457,11 +466,11 @@ pub fn print(
             } else if no_rhs_changes && !show_both {
                 match lhs_line_num {
                     Some(lhs_line_num) => {
-                        let lhs_line = &lhs_colored_lines[lhs_line_num.0];
+                        let lhs_line = &lhs_colored_lines[lhs_line_num.as_usize()];
                         if same_lines {
-                            println!("{}{}", display_lhs_line_num, lhs_line);
+                            print!("{}{}", display_lhs_line_num, lhs_line);
                         } else {
-                            println!(
+                            print!(
                                 "{}{}{}",
                                 display_lhs_line_num, display_rhs_line_num, lhs_line
                             );
@@ -474,20 +483,20 @@ pub fn print(
             } else {
                 let lhs_line = match lhs_line_num {
                     Some(lhs_line_num) => split_and_apply(
-                        lhs_lines[lhs_line_num.0],
+                        lhs_lines[lhs_line_num.as_usize()],
                         source_dims.lhs_content_width,
                         display_options.use_color,
-                        lhs_highlights.get(&lhs_line_num).unwrap_or(&vec![]),
+                        lhs_highlights.get(lhs_line_num).unwrap_or(&vec![]),
                         Side::Left,
                     ),
                     None => vec![" ".repeat(source_dims.lhs_content_width)],
                 };
                 let rhs_line = match rhs_line_num {
                     Some(rhs_line_num) => split_and_apply(
-                        rhs_lines[rhs_line_num.0],
+                        rhs_lines[rhs_line_num.as_usize()],
                         source_dims.rhs_content_width,
                         display_options.use_color,
-                        rhs_highlights.get(&rhs_line_num).unwrap_or(&vec![]),
+                        rhs_highlights.get(rhs_line_num).unwrap_or(&vec![]),
                         Side::Right,
                     ),
                     None => vec!["".into()],
@@ -511,7 +520,8 @@ pub fn print(
                             display_options.use_color,
                         );
                         if let Some(line_num) = lhs_line_num {
-                            if lhs_lines_with_novel.contains(&line_num) {
+                            if display_options.use_color && lhs_lines_with_novel.contains(line_num)
+                            {
                                 s = if display_options.background_color.is_dark() {
                                     s.bright_red().to_string()
                                 } else {
@@ -532,7 +542,8 @@ pub fn print(
                             display_options.use_color,
                         );
                         if let Some(line_num) = rhs_line_num {
-                            if rhs_lines_with_novel.contains(&line_num) {
+                            if display_options.use_color && rhs_lines_with_novel.contains(line_num)
+                            {
                                 s = if display_options.background_color.is_dark() {
                                     s.bright_green().to_string()
                                 } else {
@@ -548,10 +559,10 @@ pub fn print(
             }
 
             if lhs_line_num.is_some() {
-                prev_lhs_line_num = lhs_line_num;
+                prev_lhs_line_num = *lhs_line_num;
             }
             if rhs_line_num.is_some() {
-                prev_rhs_line_num = rhs_line_num;
+                prev_rhs_line_num = *rhs_line_num;
             }
         }
         println!();
@@ -637,42 +648,18 @@ mod tests {
         };
 
         // Basic smoke test.
-        let res = display_single_column(
+        let res_lines = display_single_column(
             "foo.py",
             "foo.py",
             "Python",
-            "print(123)\n",
+            &["print(123)\n".to_string()],
             false,
             &display_options,
         );
+        let res = res_lines.join("");
         assert!(res.len() > 10);
     }
 
-    #[test]
-    fn test_split_line_empty() {
-        assert_eq!(split_on_newlines(""), vec![""]);
-    }
-
-    #[test]
-    fn test_split_line_single() {
-        assert_eq!(split_on_newlines("foo"), vec!["foo"]);
-    }
-
-    #[test]
-    fn test_split_line_with_newline() {
-        assert_eq!(split_on_newlines("foo\nbar"), vec!["foo", "bar"]);
-    }
-
-    #[test]
-    fn test_split_line_with_crlf() {
-        assert_eq!(split_on_newlines("foo\r\nbar"), vec!["foo", "bar"]);
-    }
-
-    #[test]
-    fn test_split_line_with_trailing_newline() {
-        assert_eq!(split_on_newlines("foo\nbar\n"), vec!["foo", "bar", ""]);
-    }
-
     #[test]
     fn test_display_hunks() {
         // Simulate diffing:
